package api

import (
	"context"
	"encoding/json"
	"fmt"
	"log"
	"net/http"
	"strconv"
	"strings"
	"time"

	"golang.org/x/oauth2"

	"github.com/gorilla/mux"

	"bravo.sau.kau.se/rethink/canvas-group-lti/internal/app/cgl"
	"bravo.sau.kau.se/rethink/canvas-group-lti/internal/app/cgl/canvas"
	"bravo.sau.kau.se/rethink/canvas-group-lti/internal/app/cgl/datastore"
	"bravo.sau.kau.se/rethink/canvas-group-lti/internal/app/cgl/lti"
	"bravo.sau.kau.se/rethink/canvas-group-lti/internal/app/cgl/session"
)

// API represents the API
type API struct {
	sessionStore      *session.MemorySessionStore
	dataStore         *datastore.Datastore
	oauth2Conf        *oauth2.Config
	canvasBaseURL     string
	educationCourseID int
	tokenStoragePool  cgl.TokenStoragePool
}

// NewHandler returns a new API handler
func NewHandler(s *session.MemorySessionStore,
	ds *datastore.Datastore,
	oauth2Conf *oauth2.Config,
	canvasurl string,
	educationCourseID int) *API {

	return &API{
		sessionStore:      s,
		dataStore:         ds,
		canvasBaseURL:     canvasurl,
		oauth2Conf:        oauth2Conf,
		educationCourseID: educationCourseID,
		tokenStoragePool:  cgl.NewTokenStoragePool(),
	}
}

// ContextKey is key for context values
type ContextKey string

// SessionMW checks if API call is authorized
func (api *API) SessionMW(next http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		auth := r.Header.Get("Authorization")
		if auth == "" {
			log.Printf("Missing Authorization header. Denying access.")
			http.Error(w, "authorization failed", http.StatusUnauthorized)
			return
		}
		parts := strings.SplitN(auth, " ", 2)
		if len(parts) != 2 || parts[0] != "CGL" {
			log.Printf("Authorization header is wrong type: %v", auth)
			http.Error(w, "wrong authorization type", http.StatusUnauthorized)
			return
		}

		session := api.sessionStore.Session(parts[1])
		token, err := tokenFromSession(session)
		if err != nil {
			log.Printf("%v", err)
			http.Error(w, "invalid session", http.StatusUnauthorized)
			return
		}

		tokenStorage := api.tokenStoragePool.TokenStorageByTokenID(session.Get("tokenUserID"))
		if tokenStorage == nil {
			tokenStorage = &cgl.TokenStorage{
				New: api.oauth2Conf.TokenSource(context.Background(), &token),
				T:   &token,
				F: func(t *oauth2.Token) error {
					lti.TokenToSession(session, t.AccessToken, t.RefreshToken, t.TokenType, t.Expiry)
					return api.dataStore.TokenSave(session.Get("tokenUserID"), t.AccessToken, t.RefreshToken, t.TokenType, t.Expiry)
				},
			}
			api.tokenStoragePool.AddTokenStorage(session.Get("tokenUserID"), tokenStorage)
		}
		client := oauth2.NewClient(r.Context(), tokenStorage)
		canvasClient := canvas.NewClient(api.canvasBaseURL, client)

		courseID, _ := strconv.Atoi(session.Get("courseID"))
		userID, _ := strconv.Atoi(session.Get("userID"))
		roles := session.Get("roles")
		ctx := context.WithValue(r.Context(), ContextKey("canvasClient"), canvasClient)
		ctx = context.WithValue(ctx, ContextKey("courseID"), courseID)
		ctx = context.WithValue(ctx, ContextKey("userID"), userID)
		ctx = context.WithValue(ctx, ContextKey("roles"), roles)

		next.ServeHTTP(w, r.WithContext(ctx))
	})
}

// tokenFromSession creates an oauth2.Token from values in session
func tokenFromSession(s *session.MemorySession) (oauth2.Token, error) {
	t := oauth2.Token{
		AccessToken:  s.Get("AccessToken"),
		RefreshToken: s.Get("RefreshToken"),
		TokenType:    s.Get("TokenType"),
	}
	if t.AccessToken == "" || t.RefreshToken == "" || t.TokenType == "" {
		return t, fmt.Errorf("missing token data in session")
	}
	if err := t.Expiry.UnmarshalText([]byte(s.Get("Expiry"))); err != nil {
		return t, err
	}
	return t, nil
}

// HeaderSetterMW sets common response headers
func HeaderSetterMW(next http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		w.Header().Set("Access-Control-Allow-Origin", "*")
		w.Header().Set("Cache-Control", "no-cache, no-store, must-revalidate")
		w.Header().Set("Pragma", "no-cache")
		w.Header().Set("Expires", "0")
		next.ServeHTTP(w, r)
	})
}

// SlowerMW slows down all requests
func SlowerMW(next http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		time.Sleep(2 * time.Second)
		next.ServeHTTP(w, r)
	})
}

// RegisterRoutes register routes using the given router
func (api *API) RegisterRoutes(r *mux.Router) {
	r.Use(HeaderSetterMW)
	r.Use(api.SessionMW)
	// r.Use(SlowerMW)
	r.HandleFunc("/ltiinfo", api.LaunchInfo).Methods("GET")
	r.HandleFunc("/educationalResources", api.EducationalResources).Methods("GET")
	r.HandleFunc("/groups", api.Groups).Methods("GET")
	r.HandleFunc("/groups/{id:[0-9]+}/approvals", api.GroupsApprovals).Methods("GET")
	r.HandleFunc("/groups/{id:[0-9]+}/approvals", api.GroupsApprovalsPOST).Methods("POST").Headers("Content-Type", "application/json")
	r.HandleFunc("/groups/{id:[0-9]+}/contract", api.GroupsContract).Methods("GET")
	r.HandleFunc("/groups/{id:[0-9]+}/contract", api.GroupsContractDELETE).Methods("DELETE")
	r.HandleFunc("/groups/{id:[0-9]+}/contract", api.GroupsContractPUT).Methods("PUT").Headers("Content-Type", "application/json")
	r.HandleFunc("/groups/{id:[0-9]+}/roleDefinitions", api.GroupsRoleDefinitions).Methods("GET")
	r.HandleFunc("/groups/{id:[0-9]+}/roleDefinitions/{roledefid}", api.GroupsRoleDefinitionsDELETE).Methods("DELETE")
	r.HandleFunc("/groups/{id:[0-9]+}/roleDefinitions/{roledefid:[0-9]+}", api.GroupsRoleDefinitionsPUT).Methods("PUT").Headers("Content-Type", "application/json")
	r.HandleFunc("/groups/{id:[0-9]+}/roleDefinitions", api.GroupsRoleDefinitionsPOST).Methods("POST").Headers("Content-Type", "application/json")
	r.HandleFunc("/groups/{id:[0-9]+}/roles", api.GroupsRoles).Methods("GET")
	r.HandleFunc("/groups/{id:[0-9]+}/roles/{roleid:[0-9]+}", api.GroupsRolesDELETE).Methods("DELETE")
	r.HandleFunc("/groups/{id:[0-9]+}/roles", api.GroupsRolesPOST).Methods("POST").Headers("Content-Type", "application/json")
	r.HandleFunc("/groups/{id:[0-9]+}/users", api.GroupsUsers).Methods("GET")
	r.HandleFunc("/groups/{id:[0-9]+}/locked", api.GroupsLockedPUT).Methods("PUT").Headers("Content-Type", "application/json")
	r.HandleFunc("/groups/{id:[0-9]+}/verifications", api.GroupsVerifications).Methods("GET")
	r.HandleFunc("/groups/{id:[0-9]+}/verifications", api.GroupsVerificationsPOST).Methods("POST").Headers("Content-Type", "application/json")
	r.HandleFunc("/groups/standardContract", api.StandardContract).Methods("GET")
}

// assertUserIsGroupLeader checks that the current user is group leader.
// Used in POST, PUT and DELETE methods.
func assertUserIsGroupLeader(w http.ResponseWriter, r *http.Request, groupID int) bool {
	userID := r.Context().Value(ContextKey("userID")).(int)
	canvasClient := r.Context().Value(ContextKey("canvasClient")).(*canvas.Client)

	group, err := canvasClient.Group(groupID)
	if err != nil {
		log.Printf("API.assertUserIsGroupLeader: %v", err)
		http.Error(w, http.StatusText(http.StatusUnauthorized), http.StatusUnauthorized)
		return false
	}
	if group.GroupLeader.ID != userID {
		log.Printf("API.assertUserIsGroupLeader: user %d is not leader of group %d", userID, groupID)
		http.Error(w, http.StatusText(http.StatusUnauthorized), http.StatusUnauthorized)

		return false
	}
	return true
}

// assertUserIsInstructor checks that the current user is an instructor.
// Used in POST, PUT and DELETE methods.
func assertUserIsInstructor(w http.ResponseWriter, r *http.Request) bool {
	userID := r.Context().Value(ContextKey("userID")).(int)

	if !userIsInstructor(r) {
		log.Printf("API.assertUserIsInstructor: user %d is not an instructor", userID)
		http.Error(w, http.StatusText(http.StatusUnauthorized), http.StatusUnauthorized)
		return false
	}

	return true
}

// checks if current user is an instructor
func userIsInstructor(r *http.Request) bool {
	roles := r.Context().Value(ContextKey("roles")).(string)
	return strings.Contains(strings.ToLower(roles), "instructor")
}

func writeJSON(w http.ResponseWriter, r *http.Request, data interface{}) {
	jsondata, err := json.MarshalIndent(data, "", "  ")
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
	}
	r.Header.Set("Content-Type", "application/json")
	fmt.Fprint(w, string(jsondata))
}

// LaunchInfo represents LTI launch info from Canvas
type LaunchInfo struct {
	UserID       int    `json:"user_id"`
	CourseID     int    `json:"course_id"`
	CourseName   string `json:"course_name"`
	IsInstructor bool   `json:"is_instructor"`
}

// EducationalResourceModule represents an educational resource
type EducationalResourceModule struct {
	Name  string                    `json:"name,omitempty"`
	Items []EducationalResourceItem `json:"items,omitempty"`
}

// EducationalResourceItem represents an item in a Canvas module
type EducationalResourceItem struct {
	Title string `json:"title,omitempty"`
	Type  string `json:"type,omitempty"`
	URL   string `json:"url,omitempty"`
}

// Approval represents an approval
type Approval struct {
	UserID      int       `json:"user_id,omitempty"`
	ApproveDate time.Time `json:"approve_date,omitempty"`
}

// Contract represents a contract
type Contract struct {
	Body string `json:"body,omitempty"`
}

// Group represents a Canvas group
type Group struct {
	ID              int    `json:"id,omitempty"`
	Name            string `json:"name,omitempty"`
	GroupLeaderID   int    `json:"group_leader_id,omitempty"`
	GroupLeaderName string `json:"group_leader_name,omitempty"`
	Locked          bool   `json:"locked,omitempty"`
}

// Role represents a roledefinition mapping to a user
type Role struct {
	ID               int `json:"id,omitempty"`
	UserID           int `json:"user_id,omitempty"`
	RoleDefinitionID int `json:"role_definition_id,omitempty"`
}

// RoleDefinition represents a role definition
type RoleDefinition struct {
	ID     int    `json:"id,omitempty"`
	Name   string `json:"name,omitempty"`
	Desc   string `json:"desc,omitempty"`
	Weight int    `json:"weight,omitempty"`
}

func (rd RoleDefinition) isValid() bool {

	if len(rd.Name) == 0 || len(rd.Name) > 255 {
		return false
	}

	if len(rd.Desc) > 65535 {
		return false
	}

	if rd.Weight == 0 {
		return false
	}

	return true
}

// User represents a Canvas user
type User struct {
	ID   int    `json:"id,omitempty"`
	Name string `json:"name,omitempty"`
}

// Verification represents a teachers verification
type Verification struct {
	UserID     int       `json:"user_id,omitempty"`
	UserName   string    `json:"user_name,omitempty"`
	VerifyDate time.Time `json:"verify_date,omitempty"`
}

// LaunchInfo returns LTI informations
func (api *API) LaunchInfo(w http.ResponseWriter, r *http.Request) {
	courseID := r.Context().Value(ContextKey("courseID")).(int)
	userID := r.Context().Value(ContextKey("userID")).(int)

	ret := LaunchInfo{
		UserID:       userID,
		CourseID:     courseID,
		IsInstructor: userIsInstructor(r),
	}

	writeJSON(w, r, ret)
}

// EducationalResources returns all defined educational resources
func (api *API) EducationalResources(w http.ResponseWriter, r *http.Request) {

	// return empty array if course with educational resources not set
	if api.educationCourseID == 0 {
		writeJSON(w, r, []EducationalResourceModule{})
		return
	}

	canvasClient, ok := r.Context().Value(ContextKey("canvasClient")).(*canvas.Client)
	if !ok {
		log.Print("API.EducationalResources: no Canvas client in context")
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}

	modules, err := canvasClient.CourseModules(api.educationCourseID)
	if err != nil {
		log.Printf("API.EducationalResources: %v", err)
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}

	var resourceModules []EducationalResourceModule
	for _, module := range modules {
		resourceModule := EducationalResourceModule{Name: module.Name}
		for _, item := range module.Items {
			resourceModule.Items = append(resourceModule.Items, EducationalResourceItem{
				Title: item.Title,
				Type:  item.Type,
				URL:   item.URL,
			})
		}
		resourceModules = append(resourceModules, resourceModule)
	}

	writeJSON(w, r, resourceModules)
}

// Groups returns alla groups for the given user
func (api *API) Groups(w http.ResponseWriter, r *http.Request) {
	canvasClient, ok := r.Context().Value(ContextKey("canvasClient")).(*canvas.Client)
	if !ok {
		log.Print("API.Groups: no Canvas client in context")
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}
	courseID, ok := r.Context().Value(ContextKey("courseID")).(int)
	if !ok {
		log.Print("API.Groups: invalid courseID")
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}

	data, err := canvasClient.CourseGroups(canvas.CourseID(courseID), userIsInstructor(r))
	if err != nil {
		log.Printf("API.Groups: %v", err)
		http.Error(w, "failed to load groups", http.StatusInternalServerError)
		return
	}

	var groups = []Group{}
	for _, item := range data {
		group := Group{
			ID:              item.ID,
			Name:            item.Name,
			GroupLeaderID:   item.GroupLeader.ID,
			GroupLeaderName: item.GroupLeader.Name,
		}
		if item.GroupCategoryName != "" {
			group.Name = fmt.Sprintf("%s: %s", item.GroupCategoryName, item.Name)
		}

		// get locked status
		g, err := api.dataStore.GroupByID(group.ID)
		if err != nil {
			if err == datastore.ErrNoResult {
				group.Locked = false
			} else {
				log.Printf("API.Groups: %v", err)
				http.Error(w, "failed to load group locked status", http.StatusInternalServerError)
				return
			}
		} else {
			group.Locked = g.Locked
		}

		groups = append(groups, group)
	}

	writeJSON(w, r, groups)
}

// GroupsApprovals returns all approvals for a given group
func (api *API) GroupsApprovals(w http.ResponseWriter, r *http.Request) {
	vars := mux.Vars(r)
	groupID, _ := strconv.Atoi(vars["id"])

	data, err := api.dataStore.ApprovalsByGroupID(groupID)
	if err != nil {
		if err != datastore.ErrNoResult {
			http.Error(w, err.Error(), http.StatusInternalServerError)
			return
		}
	}

	var approvals = []Approval{}
	for _, item := range data {
		approvals = append(approvals, Approval{
			UserID:      item.UserID,
			ApproveDate: item.ApproveDate,
		})
	}

	writeJSON(w, r, approvals)
}

// GroupsApprovalsPOST sets approval for the given group and user
func (api *API) GroupsApprovalsPOST(w http.ResponseWriter, r *http.Request) {
	vars := mux.Vars(r)
	groupID, _ := strconv.Atoi(vars["id"])

	defer r.Body.Close()
	var approval Approval
	if err := json.NewDecoder(r.Body).Decode(&approval); err != nil {
		http.Error(w, fmt.Sprintf("failed to decode payload: %v", err), http.StatusBadRequest)
		return
	}

	ctxUserID := r.Context().Value(ContextKey("userID")).(int)
	if approval.UserID != ctxUserID {
		log.Printf("API.GroupsApprovalsPOST: user %d tried to create approval for user %d", ctxUserID, approval.UserID)
		http.Error(w, http.StatusText(http.StatusUnauthorized), http.StatusUnauthorized)
		return
	}

	approveDate, err := api.dataStore.ApprovalSave(groupID, approval.UserID)

	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}

	approval.ApproveDate = approveDate

	w.WriteHeader(http.StatusCreated)

	if err := json.NewEncoder(w).Encode(approval); err != nil {
		log.Printf("failed to encode JSON response: %v", err)
	}
}

// GroupsContract returns the groups contract
func (api *API) GroupsContract(w http.ResponseWriter, r *http.Request) {
	vars := mux.Vars(r)
	groupID, _ := strconv.Atoi(vars["id"])

	data, err := api.dataStore.ContractByGroupID(groupID)
	if err != nil {
		if err == datastore.ErrNoResult {
			w.WriteHeader(http.StatusNoContent)
			return
		}
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}

	contract := Contract{
		Body: data.Body,
	}

	writeJSON(w, r, contract)
}

// GroupsContractDELETE deletes the contract for the given group
func (api *API) GroupsContractDELETE(w http.ResponseWriter, r *http.Request) {
	vars := mux.Vars(r)
	groupID, _ := strconv.Atoi(vars["id"])

	if !assertUserIsGroupLeader(w, r, groupID) {
		return
	}

	if err := api.dataStore.ContractDelete(groupID); err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}

	w.WriteHeader(http.StatusNoContent)
}

// GroupsContractPUT creates or updates the contract for the given group
func (api *API) GroupsContractPUT(w http.ResponseWriter, r *http.Request) {
	vars := mux.Vars(r)
	groupID, _ := strconv.Atoi(vars["id"])

	if !assertUserIsGroupLeader(w, r, groupID) {
		return
	}

	decoder := json.NewDecoder(r.Body)
	defer r.Body.Close()
	var contract Contract
	if err := decoder.Decode(&contract); err != nil {
		errstr := fmt.Sprintf("failed to decode payload: %v", err)
		http.Error(w, errstr, http.StatusBadRequest)
		return
	}

	if err := api.dataStore.ContractSave(groupID, contract.Body); err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}

	w.WriteHeader(http.StatusNoContent)
}

// GroupsRoleDefinitions returns all role definitions for the given group
func (api *API) GroupsRoleDefinitions(w http.ResponseWriter, r *http.Request) {
	vars := mux.Vars(r)
	groupID, _ := strconv.Atoi(vars["id"])

	data, err := api.dataStore.RoleDefinitionsByGroupID(groupID)
	if err != nil {
		if err == datastore.ErrNoResult {
			http.NotFound(w, r)
			return
		}
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}

	var roledefs = []RoleDefinition{}
	for _, item := range data {
		roledefs = append(roledefs, RoleDefinition{
			ID:     item.ID,
			Name:   item.Name,
			Desc:   item.Description,
			Weight: item.Weight,
		})
	}

	writeJSON(w, r, roledefs)
}

// GroupsRoleDefinitionsDELETE deletes the role definition for the given group
func (api *API) GroupsRoleDefinitionsDELETE(w http.ResponseWriter, r *http.Request) {
	vars := mux.Vars(r)
	groupID, _ := strconv.Atoi(vars["id"])
	roleDefinitionID, _ := strconv.Atoi(vars["roledefid"])

	if !assertUserIsGroupLeader(w, r, groupID) {
		return
	}

	if err := api.dataStore.RoleDefinitionDelete(groupID, roleDefinitionID); err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}
	w.WriteHeader(http.StatusNoContent)
}

// GroupsRoleDefinitionsPOST creates a role definition for the given group
func (api *API) GroupsRoleDefinitionsPOST(w http.ResponseWriter, r *http.Request) {
	vars := mux.Vars(r)
	groupID, _ := strconv.Atoi(vars["id"])

	if !assertUserIsGroupLeader(w, r, groupID) {
		return
	}

	decoder := json.NewDecoder(r.Body)
	defer r.Body.Close()
	var rd RoleDefinition
	if err := decoder.Decode(&rd); err != nil {
		errstr := fmt.Sprintf("failed to decode payload: %v", err)
		http.Error(w, errstr, http.StatusBadRequest)
		return
	}

	if rd.isValid() == false {
		errstr := fmt.Sprintf("role definition is not valid")
		http.Error(w, errstr, http.StatusBadRequest)
		return
	}

	id, err := api.dataStore.RoleDefinitionSave(groupID, rd.Name, rd.Desc, rd.Weight)
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}

	rd.ID = int(id)

	w.WriteHeader(http.StatusCreated)

	if err := json.NewEncoder(w).Encode(rd); err != nil {
		log.Printf("failed to encode JSON response: %v", err)
	}
}

// GroupsRoleDefinitionsPUT updates a role definition
func (api *API) GroupsRoleDefinitionsPUT(w http.ResponseWriter, r *http.Request) {
	vars := mux.Vars(r)
	groupID, _ := strconv.Atoi(vars["id"])
	roleDefinitionID, _ := strconv.Atoi(vars["roledefid"])

	if !assertUserIsGroupLeader(w, r, groupID) {
		return
	}

	decoder := json.NewDecoder(r.Body)
	defer r.Body.Close()
	var rd RoleDefinition
	if err := decoder.Decode(&rd); err != nil {
		errstr := fmt.Sprintf("failed to decode payload: %v", err)
		http.Error(w, errstr, http.StatusInternalServerError)
		return
	}

	if rd.ID != roleDefinitionID {
		errstr := fmt.Sprintf("id in payload does not match URL: %d != %d", rd.ID, roleDefinitionID)
		http.Error(w, errstr, http.StatusBadRequest)
		return
	}

	if rd.isValid() == false {
		errstr := fmt.Sprintf("role definition is not valid")
		http.Error(w, errstr, http.StatusBadRequest)
		return
	}

	if err := api.dataStore.RoleDefinitionUpdate(roleDefinitionID, groupID, rd.Name, rd.Desc, rd.Weight); err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}

	w.WriteHeader(http.StatusNoContent)
}

// GroupsRoles returns all roles for the given group
func (api *API) GroupsRoles(w http.ResponseWriter, r *http.Request) {
	vars := mux.Vars(r)
	groupID, _ := strconv.Atoi(vars["id"])

	data, err := api.dataStore.RolesByGroupID(groupID)
	if err != nil {
		if err == datastore.ErrNoResult {
			http.NotFound(w, r)
			return
		}
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}

	var roles = []Role{}
	for _, item := range data {
		roles = append(roles, Role{
			ID:               int(item.ID),
			UserID:           item.UserID,
			RoleDefinitionID: item.RoleDefinitionID,
		})
	}

	writeJSON(w, r, roles)
}

// GroupsRolesDELETE deletes a role for the given group
func (api *API) GroupsRolesDELETE(w http.ResponseWriter, r *http.Request) {
	vars := mux.Vars(r)
	groupID, _ := strconv.Atoi(vars["id"])
	roleID, _ := strconv.Atoi(vars["roleid"])

	if !assertUserIsGroupLeader(w, r, groupID) {
		return
	}

	if err := api.dataStore.RolesDelete(groupID, roleID); err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}

	w.WriteHeader(http.StatusNoContent)
}

// GroupsRolesPOST creates a role mapping for the given group
func (api *API) GroupsRolesPOST(w http.ResponseWriter, r *http.Request) {
	vars := mux.Vars(r)
	groupID, _ := strconv.Atoi(vars["id"])

	if !assertUserIsGroupLeader(w, r, groupID) {
		return
	}

	decoder := json.NewDecoder(r.Body)
	defer r.Body.Close()
	var role Role
	if err := decoder.Decode(&role); err != nil {
		errstr := fmt.Sprintf("failed to decode payload: %v", err)
		http.Error(w, errstr, http.StatusBadRequest)
		return
	}

	id, err := api.dataStore.RolesSave(groupID, role.RoleDefinitionID, role.UserID)
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}

	role.ID = int(id)

	w.WriteHeader(http.StatusCreated)

	if err := json.NewEncoder(w).Encode(role); err != nil {
		log.Printf("failed to encode JSON response: %v", err)
	}
}

// GroupsUsers returns all users for the given group
func (api *API) GroupsUsers(w http.ResponseWriter, r *http.Request) {
	vars := mux.Vars(r)
	groupID, _ := strconv.Atoi(vars["id"])

	canvasClient, ok := r.Context().Value(ContextKey("canvasClient")).(*canvas.Client)
	if !ok {
		log.Print("API.Groups: no Canvas client in context")
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}

	users, err := canvasClient.GroupUsers(groupID)
	if err != nil {
		log.Printf("API.GroupsUsers: %v", err)
		http.Error(w, "error fetching users from Canvas", http.StatusInternalServerError)
		return
	}

	writeJSON(w, r, users)
}

// GroupsLockedPUT updates the locked state for the given group
// On unlock, also delete any approvals and verified state for the group.
func (api *API) GroupsLockedPUT(w http.ResponseWriter, r *http.Request) {
	vars := mux.Vars(r)
	groupID, _ := strconv.Atoi(vars["id"])

	if !assertUserIsGroupLeader(w, r, groupID) {
		return
	}

	defer r.Body.Close()
	var locked bool
	if err := json.NewDecoder(r.Body).Decode(&locked); err != nil {
		http.Error(w, fmt.Sprintf("failed to decode payload: %v", err), http.StatusBadRequest)
		return
	}

	if err := api.dataStore.GroupSave(groupID, locked); err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}

	w.WriteHeader(http.StatusNoContent)
}

// GroupsVerifications returns all verifications for a given group
func (api *API) GroupsVerifications(w http.ResponseWriter, r *http.Request) {
	vars := mux.Vars(r)
	groupID, _ := strconv.Atoi(vars["id"])

	data, err := api.dataStore.VerificationsByGroupID(groupID)
	if err != nil {
		if err != datastore.ErrNoResult {
			http.Error(w, err.Error(), http.StatusInternalServerError)
			return
		}
	}

	var verifications = []Verification{}
	for _, item := range data {
		verification := Verification{
			UserID:     item.UserID,
			UserName:   "teacher",
			VerifyDate: item.VerifyDate,
		}

		// get name of user from Canvas (if possible)
		canvasClient, ok := r.Context().Value(ContextKey("canvasClient")).(*canvas.Client)
		if ok {
			user, err := canvasClient.UserByID(verification.UserID)
			if err == nil {
				verification.UserName = user.Name
			}
		}

		verifications = append(verifications, verification)
	}

	writeJSON(w, r, verifications)
}

// GroupsVerificationsPOST sets verification for the given group and user
func (api *API) GroupsVerificationsPOST(w http.ResponseWriter, r *http.Request) {
	vars := mux.Vars(r)
	groupID, _ := strconv.Atoi(vars["id"])

	if !userIsInstructor(r) {
		return
	}
	defer r.Body.Close()
	var verification Verification
	if err := json.NewDecoder(r.Body).Decode(&verification); err != nil {
		http.Error(w, fmt.Sprintf("failed to decode payload: %v", err), http.StatusBadRequest)
		return
	}

	ctxUserID := r.Context().Value(ContextKey("userID")).(int)
	if verification.UserID != ctxUserID {
		log.Printf("API.GroupsVerificationPOST: user %d tried to create verification for user %d", ctxUserID, verification.UserID)
		http.Error(w, http.StatusText(http.StatusUnauthorized), http.StatusUnauthorized)
		return
	}

	verifyDate, err := api.dataStore.VerificationSave(groupID, verification.UserID)

	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}

	verification.VerifyDate = verifyDate

	// get name of user from Canvas (if possible)
	verification.UserName = "teacher"
	canvasClient, ok := r.Context().Value(ContextKey("canvasClient")).(*canvas.Client)
	if ok {
		user, err := canvasClient.UserByID(verification.UserID)
		if err == nil {
			verification.UserName = user.Name
		}
	}

	w.WriteHeader(http.StatusCreated)

	if err := json.NewEncoder(w).Encode(verification); err != nil {
		log.Printf("failed to encode JSON response: %v", err)
	}
}

// StandardContract returns the current standard contract
func (api *API) StandardContract(w http.ResponseWriter, r *http.Request) {
	data, err := api.dataStore.StandardContractLatest()
	if err != nil {
		if err == datastore.ErrNoResult {
			log.Printf("StandardContract: %v", err)
			http.NotFound(w, r)
			return
		}
		log.Printf("StandardContract: %v", err)
		http.Error(w, "failed to load standard contract", http.StatusInternalServerError)
		return
	}

	contract := Contract{
		Body: data.Body,
	}

	writeJSON(w, r, contract)
}
