package canvas

import (
	"encoding/json"
	"fmt"
	"log"
)

// GroupLeader represents a Canvas user who is the leader of a group
type GroupLeader struct {
	ID   int    `json:"id"`
	Name string `json:"display_name"`
}

// Group represents a Canvas group
// https://canvas.instructure.com/doc/api/groups.html#Group
type Group struct {
	ID                int         `json:"id"`
	Name              string      `json:"name"`
	GroupCategoryID   int         `json:"group_category_id"`
	GroupCategoryName string      `json:"group_category_name"`
	GroupLeader       GroupLeader `json:"leader"`
}

// CourseGroups returns the groups for a given course
// https://canvas.instructure.com/doc/api/groups.html#method.groups.context_index
func (c Client) CourseGroups(ID CourseID, isInstructor bool) ([]Group, error) {
	groups := []Group{}

	url := fmt.Sprintf("%s/courses/%v/groups?per_page=100", c.baseurl, ID)
	if !isInstructor {
		url = fmt.Sprintf("%s&only_own_groups=true", url)
	}
	log.Printf("Canvas.CourseGroups: %v", url)
	resp, err := c.client.Get(url)
	if err != nil {
		return groups, err
	}
	defer resp.Body.Close()

	if resp.StatusCode < 200 || resp.StatusCode > 299 {
		return groups, fmt.Errorf("Canvas.CourseGroups: Canvas said %v", resp.Status)
	}

	if err := json.NewDecoder(resp.Body).Decode(&groups); err != nil {
		return groups, fmt.Errorf("Canvas.CourseGroups: %v", err)
	}

	return groups, nil
}

// Group returns a specific group
// https://canvas.instructure.com/doc/api/groups.html#Group
func (c Client) Group(groupID int) (Group, error) {
	var group Group

	url := fmt.Sprintf("%s/groups/%d", c.baseurl, groupID)
	log.Printf("Canvas.Group: %s", url)
	resp, err := c.client.Get(url)
	if err != nil {
		return group, fmt.Errorf("Canvas.Group: %v", err)
	}
	defer resp.Body.Close()

	if resp.StatusCode < 200 || resp.StatusCode > 299 {
		return group, fmt.Errorf("Canvas.Group: Canvas said %v", resp.Status)
	}

	if err := json.NewDecoder(resp.Body).Decode(&group); err != nil {
		return group, fmt.Errorf("Canvas.Group: %v", err)
	}

	return group, nil
}

// GroupUsers returns the users in a given group
// https://canvas.instructure.com/doc/api/groups.html#method.groups.users
func (c Client) GroupUsers(groupID int) ([]User, error) {
	users := []User{}

	url := fmt.Sprintf("%s/groups/%d/users", c.baseurl, groupID)
	log.Printf("Canvas.GroupUsers: %v", url)
	resp, err := c.client.Get(url)
	if err != nil {
		return users, err
	}
	defer resp.Body.Close()

	if resp.StatusCode < 200 || resp.StatusCode > 299 {
		return users, fmt.Errorf("Canvas.GroupUsers: Canvas said %v", resp.Status)
	}

	if err := json.NewDecoder(resp.Body).Decode(&users); err != nil {
		return users, fmt.Errorf("Canvas.GroupUsers: %v", err)
	}

	return users, nil
}
