package canvas

import (
	"encoding/json"
	"fmt"
	"log"
)

// Module represents a Canvas module
type Module struct {
	ID    int          `json:"id,omitempty"`
	Name  string       `json:"name,omitempty"`
	Items []ModuleItem `json:"items,omitempty"`
}

// ModuleItem represents a Canvas module item
type ModuleItem struct {
	ID    int    `json:"id,omitempty"`
	Title string `json:"title,omitempty"`
	Type  string `json:"type,omitempty"`
	URL   string `json:"html_url,omitempty"`
}

// CourseModules returns the modules for a given course
func (c Client) CourseModules(courseID int) ([]Module, error) {
	modules := []Module{}

	url := fmt.Sprintf("%s/courses/%v/modules?include=items", c.baseurl, courseID)
	log.Printf("Canvas.CourseModules: %v", url)
	resp, err := c.client.Get(url)
	if err != nil {
		return modules, err
	}
	defer resp.Body.Close()

	if resp.StatusCode < 200 || resp.StatusCode > 299 {
		return modules, fmt.Errorf("Canvas.CourseModules: Canvas said %v", resp.Status)
	}

	if err := json.NewDecoder(resp.Body).Decode(&modules); err != nil {
		return modules, fmt.Errorf("Canvas.CourseModules: %v", err)
	}

	// Fetch module items separately if previous call did not return them
	for i := range modules {
		if modules[i].Items != nil {
			continue
		}
		modules[i].Items, err = c.CourseModuleItems(courseID, modules[i].ID)
		if err != nil {
			return modules, err
		}
	}

	return modules, nil
}

// CourseModuleItems returnes the items for a given module
func (c Client) CourseModuleItems(courseID, moduleID int) ([]ModuleItem, error) {
	var items []ModuleItem

	url := fmt.Sprintf("%s/courses/%v/modules/%v/items", c.baseurl, courseID, moduleID)
	log.Printf("Canvas.CourseModules: %v", url)
	resp, err := c.client.Get(url)
	if err != nil {
		return items, err
	}
	defer resp.Body.Close()

	if resp.StatusCode < 200 || resp.StatusCode > 299 {
		return items, fmt.Errorf("Canvas.CourseModules: Canvas said %v", resp.Status)
	}

	if err := json.NewDecoder(resp.Body).Decode(&items); err != nil {
		return items, fmt.Errorf("Canvas.CourseModules: %v", err)
	}

	return items, nil
}
