package datastore

import (
	"database/sql"
	"errors"
	"fmt"
	"time"

	_ "github.com/go-sql-driver/mysql" // MySQL
)

// Datastore represents a datastore
type Datastore struct {
	db *sql.DB
}

// New returns a new datastore
func New(urn string) (*Datastore, error) {
	db, err := sql.Open("mysql", urn)
	if err != nil {
		return nil, fmt.Errorf("failed to open database: %v", err)
	}
	if err = db.Ping(); err != nil {
		return nil, err
	}
	ds := &Datastore{
		db: db,
	}
	return ds, nil
}

// Approval represents an approvals table
type Approval struct {
	ID          int
	GroupID     int
	UserID      int
	ApproveDate time.Time
}

// Contract represents a contracts table
type Contract struct {
	ID      int
	GroupID int
	Body    string
}

//Group represents a groups table
type Group struct {
	ID      int
	GroupID int
	Locked  bool
}

// RoleDefinition represents a roledefs table
type RoleDefinition struct {
	ID          int
	GroupID     int
	Name        string
	Description string
	Weight      int
}

// Role represents a roles table
type Role struct {
	ID               int
	GroupID          int
	RoleDefinitionID int
	UserID           int
}

// Verification represents a verifications table
type Verification struct {
	ID         int
	GroupID    int
	UserID     int
	VerifyDate time.Time
}

// Token represents an OAuth2 token
type Token struct {
	AccessToken  string
	RefreshToken string
	TokenType    string
	Expiry       time.Time
}

// ErrNoResult is used when a query returns no result. Corresponds to sql.ErrNoRows.
var ErrNoResult error = errors.New("no result found")

// ApprovalsByGroupID returns all approvals for the specified group
func (ds *Datastore) ApprovalsByGroupID(groupID int) ([]Approval, error) {
	var approvals []Approval

	rows, err := ds.db.Query("SELECT id,groupid,userid,approvedate FROM approvals WHERE groupid=?", groupID)
	if err != nil {
		if err == sql.ErrNoRows {
			return approvals, ErrNoResult
		}
		return approvals, fmt.Errorf("failed to fetch approvals for group %d: %v", groupID, err)
	}
	defer rows.Close()

	for rows.Next() {
		a := Approval{}
		if err := rows.Scan(&a.ID, &a.GroupID, &a.UserID, &a.ApproveDate); err != nil {
			return approvals, fmt.Errorf("failed to read approvals for group %d: %v", groupID, err)
		}
		approvals = append(approvals, a)
	}

	return approvals, nil
}

// ApprovalSave saves the approval to the database
func (ds *Datastore) ApprovalSave(groupID, userID int) (time.Time, error) {

	approveDate := time.Now()

	_, err := ds.db.Exec("INSERT INTO approvals (groupid, userid, approvedate) VALUES (?,?,?);", groupID, userID, approveDate)
	if err != nil {
		return approveDate, fmt.Errorf("failed to write approval for group %d: %v", groupID, err)
	}

	return approveDate, nil
}

// ContractByGroupID returns a contract
func (ds *Datastore) ContractByGroupID(groupID int) (Contract, error) {
	c := Contract{}

	row := ds.db.QueryRow("SELECT id,groupid,body FROM contracts WHERE groupid=?;", groupID)
	if err := row.Scan(&c.ID, &c.GroupID, &c.Body); err != nil {
		if err == sql.ErrNoRows {
			return c, ErrNoResult
		}
		return c, fmt.Errorf("failed to fetch contract for group %d: %v", groupID, err)
	}

	return c, nil
}

// ContractDelete deletes the group's contract
func (ds *Datastore) ContractDelete(groupID int) error {
	_, err := ds.db.Exec("DELETE FROM contracts WHERE groupid=?;", groupID)
	if err != nil {
		return fmt.Errorf("error while deleting contract for group %d: %v", groupID, err)
	}
	return nil
}

// ContractSave saves the contract to the database
func (ds *Datastore) ContractSave(groupID int, body string) error {
	_, err := ds.db.Exec("INSERT INTO contracts (groupid, body) VALUES(?,?) ON DUPLICATE KEY UPDATE body=VALUES(body);", groupID, body)
	if err != nil {
		return fmt.Errorf("failed to write contract for group %d: %v", groupID, err)
	}
	return nil
}

//GroupByID returns the group with the given ID
func (ds *Datastore) GroupByID(groupID int) (Group, error) {
	g := Group{}

	row := ds.db.QueryRow("SELECT id,groupid,locked FROM groups WHERE groupid=?;", groupID)
	if err := row.Scan(&g.ID, &g.GroupID, &g.Locked); err != nil {
		if err == sql.ErrNoRows {
			return g, ErrNoResult
		}
		return g, fmt.Errorf("failed to fetch group %d: %v", groupID, err)
	}

	return g, nil
}

// GroupSave saves the group to the database
// If locked is false, also remove approvals and verifications for the group.
func (ds *Datastore) GroupSave(groupID int, locked bool) error {
	tx, err := ds.db.Begin()
	if err != nil {
		return fmt.Errorf("failed to start transaction: %v", err)
	}

	_, err = tx.Exec("INSERT INTO groups (groupid, locked) VALUES(?,?) ON DUPLICATE KEY UPDATE locked=VALUES(locked);", groupID, locked)
	if err != nil {
		tx.Rollback()
		return fmt.Errorf("failed to write group %d: %v", groupID, err)
	}

	if !locked {
		_, err = tx.Exec("DELETE FROM verifications WHERE groupid=?", groupID)
		if err != nil {
			tx.Rollback()
			return fmt.Errorf("failed to remove verifications on unlock for group %d: %v", groupID, err)
		}

		_, err = tx.Exec("DELETE FROM approvals WHERE groupid=?", groupID)
		if err != nil {
			tx.Rollback()
			return fmt.Errorf("failed to remove approvals on unlock for group %d: %v", groupID, err)
		}
	}

	tx.Commit()

	return nil
}

// RolesByGroupID returns all roles for the given group
func (ds *Datastore) RolesByGroupID(groupID int) ([]Role, error) {
	var roles []Role

	rows, err := ds.db.Query("SELECT id,groupid,roledefid,userid FROM roles WHERE groupid=?;", groupID)
	if err != nil {
		if err == sql.ErrNoRows {
			return roles, ErrNoResult
		}
		return roles, fmt.Errorf("failed to fetch roles for group %d: %v", groupID, err)
	}
	defer rows.Close()

	for rows.Next() {
		r := Role{}
		if err := rows.Scan(&r.ID, &r.GroupID, &r.RoleDefinitionID, &r.UserID); err != nil {
			return roles, fmt.Errorf("failed to read roles for group %d: %v", groupID, err)
		}
		roles = append(roles, r)
	}

	return roles, nil
}

// RolesDelete deletes the role
func (ds *Datastore) RolesDelete(groupID, roleID int) error {
	_, err := ds.db.Exec("DELETE FROM roles WHERE groupid=? AND id=?;", groupID, roleID)
	if err != nil {
		return fmt.Errorf("error while deleting role for group %d: %v", groupID, roleID)
	}
	return nil
}

// RolesSave saves the role to the database
// Returns an ID or an error
func (ds *Datastore) RolesSave(groupID, roleDefID, userID int) (int64, error) {
	result, err := ds.db.Exec("INSERT INTO roles (groupid, roledefid, userid) VALUES (?,?,?);", groupID, roleDefID, userID)
	if err != nil {
		return 0, fmt.Errorf("failed to write role for group %d: %v", groupID, err)
	}

	id, err := result.LastInsertId()
	if err != nil {
		return -1, nil
	}

	return id, nil
}

// RoleDefinitionsByGroupID returns all roles för the given group
func (ds *Datastore) RoleDefinitionsByGroupID(groupID int) ([]RoleDefinition, error) {
	var roledefs []RoleDefinition

	rows, err := ds.db.Query("SELECT id,groupid,name,description,weight FROM roledefs WHERE groupid=?;", groupID)
	if err != nil {
		if err == sql.ErrNoRows {
			return roledefs, ErrNoResult
		}
		return roledefs, fmt.Errorf("failed to fetch role definitions for group %d: %v", groupID, err)
	}
	defer rows.Close()

	for rows.Next() {
		rd := RoleDefinition{}
		if err := rows.Scan(&rd.ID, &rd.GroupID, &rd.Name, &rd.Description, &rd.Weight); err != nil {
			return roledefs, fmt.Errorf("failed to read role definitions for group %d: %v", groupID, err)
		}
		roledefs = append(roledefs, rd)
	}

	return roledefs, nil
}

// RoleDefinitionDelete deletes the given role definition
func (ds *Datastore) RoleDefinitionDelete(groupID, roleDefinitionID int) error {
	_, err := ds.db.Exec("DELETE from roledefs WHERE groupid=? AND id=?;", groupID, roleDefinitionID)
	if err != nil {
		return fmt.Errorf("error while deleting role definition for group %d: %v", groupID, err)
	}
	return nil
}

// RoleDefinitionSave saves the role definition to the database.
// It returns an ID or an error.
func (ds *Datastore) RoleDefinitionSave(groupID int, name, desc string, weight int) (int64, error) {
	result, err := ds.db.Exec("INSERT INTO roledefs (groupid, name, description, weight) VALUES (?,?,?,?);",
		groupID, name, desc, weight)
	if err != nil {
		return -1, fmt.Errorf("failed to create role definition for group %d: %v", groupID, err)
	}

	id, err := result.LastInsertId()
	if err != nil {
		return -1, nil
	}

	return id, nil
}

// RoleDefinitionUpdate updates the given role definition
func (ds *Datastore) RoleDefinitionUpdate(roleDefinitionID, groupID int, name, desc string, weight int) error {
	_, err := ds.db.Exec("UPDATE roledefs SET name=?, description=?, weight=? WHERE groupid=? AND id=?;",
		name, desc, weight, groupID, roleDefinitionID)
	if err != nil {
		return fmt.Errorf("failed to update role definition %d for group %d: %v", roleDefinitionID, groupID, err)
	}
	return nil
}

// StandardContractLatest return the newest standard contract
func (ds *Datastore) StandardContractLatest() (Contract, error) {
	var c Contract

	row := ds.db.QueryRow("SELECT id,body FROM stdcontracts ORDER BY created DESC LIMIT 1")
	if err := row.Scan(&c.ID, &c.Body); err != nil {
		if err == sql.ErrNoRows {
			return c, ErrNoResult
		}
		return c, fmt.Errorf("failed to fetch standard contract: %v", err)
	}
	return c, nil
}

// Token retrieves a token
func (ds *Datastore) Token(userID string) (Token, error) {
	var t Token

	row := ds.db.QueryRow("SELECT access,refresh,tokentype,expiry FROM tokens WHERE userid=?;", userID)
	if err := row.Scan(&t.AccessToken, &t.RefreshToken, &t.TokenType, &t.Expiry); err != nil {
		if err == sql.ErrNoRows {
			return t, ErrNoResult
		}
		return t, fmt.Errorf("failed to fetch token: %v", err)
	}

	return t, nil
}

// TokenSave saves a token
func (ds *Datastore) TokenSave(userid, accessToken, refreshToken, tokenType string, expiry time.Time) error {
	_, err := ds.db.Exec("INSERT INTO tokens (userid,access,refresh,tokentype,expiry) VALUES(?,?,?,?,?) ON DUPLICATE KEY UPDATE access=VALUES(access), expiry=VALUES(expiry)", userid, accessToken, refreshToken, tokenType, expiry)
	if err != nil {
		return fmt.Errorf("failed to save token in database: %v", err)
	}

	return nil
}

// VerificationsByGroupID returns all verifications for the specified group
func (ds *Datastore) VerificationsByGroupID(groupID int) ([]Verification, error) {
	var verifications []Verification

	rows, err := ds.db.Query("SELECT id,groupid,userid,verifydate FROM verifications WHERE groupid=?", groupID)
	if err != nil {
		if err == sql.ErrNoRows {
			return verifications, ErrNoResult
		}
		return verifications, fmt.Errorf("failed to fetch verifications for group %d: %v", groupID, err)
	}
	defer rows.Close()

	for rows.Next() {
		v := Verification{}
		if err := rows.Scan(&v.ID, &v.GroupID, &v.UserID, &v.VerifyDate); err != nil {
			return verifications, fmt.Errorf("failed to read verifications for group %d: %v", groupID, err)
		}
		verifications = append(verifications, v)
	}

	return verifications, nil
}

// VerificationSave saves the verification to the database
func (ds *Datastore) VerificationSave(groupID int, userID int) (time.Time, error) {
	verifyDate := time.Now()

	_, err := ds.db.Exec("INSERT INTO verifications (groupid, userid, verifydate) VALUES (?,?,?);", groupID, userID, verifyDate)
	if err != nil {
		return verifyDate, fmt.Errorf("failed to write verification for group %d: %v", groupID, err)
	}

	return verifyDate, nil
}
