package lti

import (
	"crypto/md5"
	"fmt"
	"io/ioutil"
	"log"
	"net/http"
	"time"

	"golang.org/x/oauth2"

	"github.com/gorilla/mux"

	"bravo.sau.kau.se/rethink/canvas-group-lti/internal/app/cgl/datastore"
	"bravo.sau.kau.se/rethink/canvas-group-lti/internal/app/cgl/oauth"
	"bravo.sau.kau.se/rethink/canvas-group-lti/internal/app/cgl/session"
)

// Handler hanterar LTI-frågor
type Handler struct {
	sessionStore        *session.MemorySessionStore
	oauthConsumerKey    string
	oauthConsumerSecret string
	oauth2Conf          *oauth2.Config
	datastore           *datastore.Datastore
}

// NewHandler returnerar en ny Handler
func NewHandler(consumerKey, consumerSecret string, oauth2Config *oauth2.Config, session *session.MemorySessionStore, ds *datastore.Datastore) *Handler {
	return &Handler{
		sessionStore:        session,
		oauthConsumerKey:    consumerKey,
		oauthConsumerSecret: consumerSecret,
		oauth2Conf:          oauth2Config,
		datastore:           ds,
	}
}

// RegisterRoutes registers routes using the given router
func (h *Handler) RegisterRoutes(router *mux.Router) {
	router.HandleFunc("/launch", h.Launch).Methods("POST")
	router.HandleFunc("/checkOAuth2", h.CheckOAuth2)
}

// Launch hanterar LTI-launch
func (h *Handler) Launch(w http.ResponseWriter, r *http.Request) {
	err := r.ParseForm()
	if err != nil {
		log.Printf("could not parse form: %v", err)
		http.Error(w, "could not parse form", http.StatusBadRequest)
		return
	}
	_, err = ioutil.ReadAll(r.Body)
	if err != nil {
		log.Printf("could not read body: %v", err)
		http.Error(w, "could not read body", http.StatusBadRequest)
		return
	}
	r.Body.Close()

	if key := r.FormValue("oauth_consumer_key"); key != h.oauthConsumerKey {
		log.Printf("invalid OAuth Consumer Key: %v", key)
		http.Error(w, "invalid OAuth consumer key", http.StatusBadRequest)
		return
	}

	if msgtype := r.FormValue("lti_message_type"); msgtype != "basic-lti-launch-request" {
		log.Printf("this is not a launch: %v", msgtype)
		http.Error(w, "failed to launch LTI app", http.StatusBadRequest)
		return
	}

	if ltiversion := r.FormValue("lti_version"); ltiversion != "LTI-1p0" {
		log.Printf("wrong LTI version: LTI-1p0 != %v", ltiversion)
		http.Error(w, "wrong LTI version received", http.StatusBadRequest)
		return
	}

	if err := oauth.Verify(r, h.oauthConsumerSecret, ""); err != nil {
		log.Printf("error verifying OAuth: %v", err)
		fmt.Fprint(w, "<!DOCTYPE html><html><body><h1>Failed to launch app</h1><p>We could not validate oAuth :/</p></body></html>")
		return
	}

	customCanvasUserID := r.FormValue("custom_canvas_user_id")
	if customCanvasUserID == "" {
		log.Print("missing POST variable custom_canvas_user_id (needed for session ID)")
		http.Error(w, "missing required information from LMS", http.StatusBadRequest)
		return
	}

	customCanvasCourseID := r.FormValue("custom_canvas_course_id")
	if customCanvasCourseID == "" {
		log.Print("missing POST variable custom_canvas_course_id (needed for session ID)")
		http.Error(w, "missing required information from LMS", http.StatusBadRequest)
		return
	}

	customCanvasAPIDomain := r.FormValue("custom_canvas_api_domain")
	if customCanvasAPIDomain == "" {
		log.Print("missing POST variable custom_canvas_api_domain (needed for session ID)")
		http.Error(w, "missing required information from LMS", http.StatusBadRequest)
		return
	}

	userID := r.FormValue("user_id")
	if userID == "" {
		log.Print("missing POST variable user_id (needed for session ID)")
		http.Error(w, "missing required information from LMS", http.StatusBadRequest)
		return
	}

	roles := r.FormValue("roles")
	if roles == "" {
		log.Print("missing POST variable roles")
		http.Error(w, "missing required information from LMS", http.StatusBadRequest)
		return
	}

	// tokenUserID is md5(SECRET:custom_canvas_api_domain:user_id)
	// Used for storing the user's OAuth2 token
	hash := md5.Sum([]byte(fmt.Sprintf("%s:%s:%s", "meep", customCanvasAPIDomain, userID)))
	tokenUserID := fmt.Sprintf("%x", hash)

	// sessionID is md5(SECRET:custom_canvas_api_domain:user_id:custom_canvas_course_id)
	// This gives us separate sessions per user & course.
	hash = md5.Sum([]byte(fmt.Sprintf("%s:%s:%s:%s", "meep", customCanvasAPIDomain, userID, customCanvasCourseID)))
	sessionID := fmt.Sprintf("%x", hash)

	session := h.sessionStore.Session(sessionID)
	session.Set("tokenUserID", tokenUserID)
	session.Set("userID", customCanvasUserID)
	session.Set("courseID", customCanvasCourseID)
	session.Set("roles", roles)

	token, err := h.datastore.Token(tokenUserID)
	if err != nil {
		if err == datastore.ErrNoResult {
			// No previous OAuth2 token found. Request authorization from user.
			url := h.oauth2Conf.AuthCodeURL(sessionID, oauth2.AccessTypeOffline)
			http.Redirect(w, r, url, http.StatusSeeOther)
			return
		}
		log.Printf("could not load token from datastore: %v", err)
		http.Error(w, "could not load token", http.StatusInternalServerError)
		return
	}

	// We have a previous OAuth2 token. Save info in session.
	if err := TokenToSession(session, token.AccessToken, token.RefreshToken, token.TokenType, token.Expiry); err != nil {
		log.Print(err.Error())
		http.Error(w, "error while reading access information", http.StatusInternalServerError)
		return
	}

	// Redirect to app.
	http.Redirect(w, r, fmt.Sprintf("/?session=%s", sessionID), http.StatusSeeOther)
}

// CheckOAuth2 checks if oauth2 flow went ok
func (h *Handler) CheckOAuth2(w http.ResponseWriter, r *http.Request) {
	log.Printf("%v request from %v", r.Method, r.RemoteAddr)

	params := r.URL.Query()
	if _, ok := params["error"]; ok {
		log.Printf("Access denied")
		fmt.Fprintf(w, "Sorry, your permission is required to proceed.")
		return
	}
	code, ok := params["code"]
	if !ok {
		log.Print("No code parameter found :/")
		http.Error(w, "no code parameter found", http.StatusBadRequest)
		return
	}
	state, ok := params["state"]
	if !ok {
		log.Printf("No state parameter found :/")
		http.Error(w, "no state parameter found", http.StatusBadRequest)
		return
	}
	log.Printf("State parameter %v found.", state)

	sessionID := state[0]
	session := h.sessionStore.Session(sessionID)
	tokenUserID := session.Get("tokenUserID")
	if tokenUserID == "" {
		log.Printf("Session %v lacks key tokenUserID.", sessionID)
		http.Error(w, "no session found", http.StatusBadRequest)
		return
	}

	// Code found and session is valid. Exchanging authorization code for
	// access and refresh tokens.
	token, err := h.oauth2Conf.Exchange(r.Context(), code[0])
	if err != nil {
		log.Printf("error exchanging oauth2 token: %v", err)
		http.Error(w, "failed to exchange oauth2 token", http.StatusInternalServerError)
		return
	}

	if err := h.datastore.TokenSave(tokenUserID, token.AccessToken, token.RefreshToken, token.TokenType, token.Expiry); err != nil {
		log.Printf("Error persisting token: %v", err)
		http.Error(w, "error saving permission", http.StatusInternalServerError)
		return
	}

	if err := TokenToSession(session, token.AccessToken, token.RefreshToken, token.TokenType, token.Expiry); err != nil {
		log.Printf("%v", err)
		http.Error(w, "error while reading access information", http.StatusInternalServerError)
		return
	}

	log.Printf("Token received: %+v", token)
	log.Print("Redirecting to application.")

	http.Redirect(w, r, fmt.Sprintf("/?session=%s", sessionID), http.StatusSeeOther)
}

// TokenToSession saves token information to session
func TokenToSession(s *session.MemorySession, accessToken, refreshToken, tokenType string, expiry time.Time) error {
	s.Set("AccessToken", accessToken)
	s.Set("RefreshToken", refreshToken)
	s.Set("TokenType", tokenType)
	expiryStr, err := expiry.MarshalText()
	if err != nil {
		return fmt.Errorf("could not convert expiry time to string: %v", err)
	}
	s.Set("Expiry", string(expiryStr))
	return nil
}
