package oauth

import (
	"crypto/hmac"
	"crypto/sha1"
	"encoding/base64"
	"errors"
	"fmt"
	"net/http"
	"sort"
	"strconv"
	"strings"
	"time"
)

// Algorithm from
// https://developer.twitter.com/en/docs/basics/authentication/guides/creating-a-signature.html
// and https://tools.ietf.org/html/rfc5849

var nonces map[string]bool

// Verify returns an error if OAuth verification fails
func Verify(r *http.Request, consumerSecret, tokenSecret string) error {
	if nonces == nil {
		nonces = map[string]bool{}
	}
	// Check OAuth nonce
	nonce := r.FormValue("oauth_nonce")

	if nonce == "" {
		return errors.New("oauth nonce is missing")
	}
	if _, ok := nonces[nonce]; ok {
		return errors.New("oauth nonce has already been seen")
	}
	nonces[nonce] = true

	// Check OAuth timestamp
	timestamp := r.FormValue("oauth_timestamp")
	if timestamp == "" {
		return errors.New("oauth timestamp is missing")
	}
	ts, err := strconv.Atoi(timestamp)
	if err != nil {
		return fmt.Errorf("oauth timestamp must be an integer value: %v", err)
	}

	now := time.Now().Unix()
	diff := now - int64(ts)
	if diff < -60 {
		return fmt.Errorf("oauth timestamp is from the future: now %v, ts %v, diff %v", now, int64(ts), diff)
	}
	if diff > 60*5 {
		return fmt.Errorf("oauth timestamp is older than five minutes (%v seconds)", diff)
	}

	// Check OAuth request signature

	expectedSignature, err := generateOAuthSignature(r, consumerSecret, tokenSecret)
	if err != nil {
		return err
	}

	if r.FormValue("oauth_signature") != expectedSignature {
		return fmt.Errorf("oauth signature does not match, %v != %v", r.FormValue("oauth_signature"), expectedSignature)
	}

	return nil
}

// GenerateOAuthSignature generates an OAuth signature for the request
func generateOAuthSignature(r *http.Request, consumerSecret, tokenSecret string) (string, error) {
	signatureBaseString, err := generateSignatureBaseString(r)
	if err != nil {
		return "", fmt.Errorf("error generating signature base string: %v", err)
	}

	signingKey := generateSigningKey(consumerSecret, tokenSecret)

	// Hash with HMAC-SHA1
	mac := hmac.New(sha1.New, []byte(signingKey))
	mac.Write([]byte(signatureBaseString))

	// Convert data to Base64 encoded string
	hashBase64 := base64.StdEncoding.EncodeToString(mac.Sum(nil))

	return hashBase64, nil
}

// https://tools.ietf.org/html/rfc5849#section-3.4.1
func generateSignatureBaseString(r *http.Request) (string, error) {
	// Parameter values must be sorted alphabetically by key
	var keys []string
	for key := range r.Form {
		if key == "oauth_signature" {
			continue
		}
		keys = append(keys, key)
	}
	sort.Strings(keys)

	var parameters string
	for _, key := range keys {
		if parameters != "" {
			parameters += "&"
		}
		// log.Printf("  %v => %v", key, r.Form.Get(key))
		parameters += percentEncode(key) + "=" + percentEncode(r.Form.Get(key))
	}

	method := strings.ToUpper(r.Method)

	protocol := r.Header.Get("X-Forwarded-Proto")
	if protocol == "" {
		protocol = "https"
	}

	hostname := r.Header.Get("X-Appengine-Default-Version-Hostname")
	if hostname == "" {
		hostname = r.Host
	}

	urlstr := fmt.Sprintf("%v://%v%v", protocol, hostname, r.URL.Path)
	sigbasestr := fmt.Sprintf("%v&%v&%v", method, percentEncode(urlstr), percentEncode(parameters))

	return sigbasestr, nil
}

func generateSigningKey(consumerSecret, tokenSecret string) string {
	return percentEncode(consumerSecret) + "&" + percentEncode(tokenSecret)
}

// https://tools.ietf.org/html/rfc5849#section-3.6
func percentEncode(in string) string {
	doNotEncode := "abcdefghijklmnopqrstuvwxyz"
	doNotEncode += "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
	doNotEncode += "0123456789"
	doNotEncode += "-._~"

	var out strings.Builder
	for i := 0; i < len(in); i++ {
		if strings.Contains(doNotEncode, string(in[i])) {
			out.WriteByte(in[i])
		} else {
			encoded := fmt.Sprintf("%%%X", in[i])
			out.WriteString(encoded)
		}
	}
	return out.String()
}
