package cgl

import (
	"sync"

	"golang.org/x/oauth2"
)

// NewTokenFunc is called for new access tokens
type NewTokenFunc func(*oauth2.Token) error

// TokenStorage stores tokens
type TokenStorage struct {
	sync.Mutex
	New oauth2.TokenSource
	T   *oauth2.Token
	F   NewTokenFunc
}

// Token returns the current token if valid
func (ts *TokenStorage) Token() (*oauth2.Token, error) {
	ts.Lock()
	defer ts.Unlock()

	if ts.T.Valid() {
		return ts.T, nil
	}

	t, err := ts.New.Token()
	if err != nil {
		return nil, err
	}

	ts.T = t
	return t, ts.F(t)
}

// TokenStoragePool stores TokenStorage objects
type TokenStoragePool struct {
	sync.Mutex
	pool map[string]*TokenStorage
}

// NewTokenStoragePool returns a new TokenStoragePool
func NewTokenStoragePool() TokenStoragePool {
	return TokenStoragePool{
		pool: make(map[string]*TokenStorage),
	}
}

// TokenStorageByTokenID returns a TokenStorage for the given token ID
func (pool *TokenStoragePool) TokenStorageByTokenID(tokenID string) *TokenStorage {
	pool.Lock()
	defer pool.Unlock()

	ts, ok := pool.pool[tokenID]
	if !ok {
		return nil
	}

	return ts
}

// AddTokenStorage adds a TokenStorage to the pool
func (pool *TokenStoragePool) AddTokenStorage(tokenID string, ts *TokenStorage) {
	pool.Lock()
	defer pool.Unlock()

	pool.pool[tokenID] = ts
}
