package main

import (
	"log"
	"net/http"
	"time"

	"golang.org/x/oauth2"

	"github.com/gorilla/mux"

	"bravo.sau.kau.se/rethink/canvas-group-lti/internal/app/cgl"
	"bravo.sau.kau.se/rethink/canvas-group-lti/internal/app/cgl/api"
	"bravo.sau.kau.se/rethink/canvas-group-lti/internal/app/cgl/datastore"
	"bravo.sau.kau.se/rethink/canvas-group-lti/internal/app/cgl/lti"
	"bravo.sau.kau.se/rethink/canvas-group-lti/internal/app/cgl/session"
)

// LoggerMW logs all requests
func LoggerMW(next http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		start := time.Now().UTC()
		next.ServeHTTP(w, r)
		diff := time.Now().UTC().Sub(start)
		uniqueid := r.Header.Get("X-Unique-ID")
		remoteAddr := r.Header.Get("X-Forwarded-For")
		if remoteAddr == "" {
			remoteAddr = r.RemoteAddr
		}
		log.Printf("%s %s %s %s %.5fs %s", remoteAddr, r.Method, r.RequestURI, r.Proto, diff.Seconds(), uniqueid)
	})
}

func main() {
	conf, err := cgl.ReadConfig("cgl.json")
	if err != nil {
		log.Fatalf("error reading config file: %v", err)
	}

	dataStore, err := datastore.New(conf.DB.URN)
	if err != nil {
		log.Fatalf("failed to open datastore: %v", err)
	}
	oauth2Conf := &oauth2.Config{
		ClientID:     conf.Canvas.ClientID,
		ClientSecret: conf.Canvas.ClientSecret,
		Endpoint: oauth2.Endpoint{
			AuthURL:  conf.Canvas.AuthURL,
			TokenURL: conf.Canvas.TokenURL,
		},
		Scopes: []string{
			"/auth/userinfo",
			"url:GET|/api/v1/courses/:course_id/groups",
			"url:GET|/api/v1/courses/:course_id/group_categories",
			"url:GET|/api/v1/courses/:course_id/modules",
			"url:GET|/api/v1/courses/:course_id/modules/:module_id/items",
			"url:GET|/api/v1/groups/:group_id/users",
			"url:GET|/api/v1/groups/:group_id",
			"url:GET|/api/v1/users/:id",
		},
		RedirectURL: conf.Canvas.RedirectURL,
	}
	sessionStore := session.NewMemorySessionStore()

	r := mux.NewRouter()
	r.Use(LoggerMW)
	lti.NewHandler(
		conf.OAuth.ConsumerKey,
		conf.OAuth.ConsumerSecret,
		oauth2Conf,
		sessionStore,
		dataStore,
	).RegisterRoutes(r.PathPrefix("/lti/").Subrouter())
	api.NewHandler(
		sessionStore,
		dataStore,
		oauth2Conf,
		conf.Canvas.BaseURL,
		conf.EducationalResources.CourseID,
	).RegisterRoutes(r.PathPrefix("/api/").Subrouter())
	r.PathPrefix("/").Handler(http.StripPrefix("", http.FileServer(http.Dir("web/static"))))
	http.Handle("/", r)

	if err := http.ListenAndServeTLS(":443", conf.HTTPS.CertPath, conf.HTTPS.KeyPath, nil); err != nil {
		log.Fatal(err)
	}
}
