import { Component, ElementRef, Inject, ViewChild, ChangeDetectorRef } from '@angular/core';

import { BehaviorSubject } from 'rxjs';

import { NgbTabset, NgbTabChangeEvent } from '@ng-bootstrap/ng-bootstrap';

import { Group } from './shared/models/group';
import { LaunchInfo } from './shared/models/launch-info';
import { GroupService } from './shared/services/group.service';
import { BusyService } from './shared/services/busy.service';

import { ConfirmDeleteModalComponent } from './confirm-delete-modal/confirm-delete-modal.component';
import { ContractComponent } from './contract/contract.component';
import { RoleDefinitionsComponent } from './role-definitions/role-definitions.component';
import { RolesComponent } from './roles/roles.component';
import { ApprovalComponent } from './approval/approval.component';

@Component({
  selector: 'app-root',
  templateUrl: './app.component.html',
  styleUrls: ['./app.component.css'],
})
export class AppComponent {
  title = 'canvas-group-lti-ui';
  readonly: boolean;
  group: Group;
  launchInfo: LaunchInfo;

  tabRequest: string;

  @ViewChild('confirmDeleteModal') confirmDeleteModal: ConfirmDeleteModalComponent;
  @ViewChild('tabs') tabs: NgbTabset;

  @ViewChild('contract') contract: ContractComponent;
  @ViewChild('roleDefinitions') roleDefinitions: RoleDefinitionsComponent;
  @ViewChild('roles') roles: RolesComponent;
  @ViewChild('approvals') approvals: ApprovalComponent;

  isBusy = false;

  constructor(
    private busyService: BusyService,
    private elementRef: ElementRef,
    private groupService: GroupService,
    private changeDetectionRef: ChangeDetectorRef,
    @Inject('CGLSessionToken') sessionToken: BehaviorSubject<string>
  ) {
    // sessionToken is an attribute from index.html:app-root.
    sessionToken.next(this.elementRef.nativeElement.getAttribute('sessionToken'));

    groupService.getLaunchInfo()
    .subscribe(result => this.launchInfo = result);
  }

  ngAfterViewInit() {
    this.busyService.isBusy
      .subscribe((isBusy: boolean) => {
          this.isBusy = isBusy;
          this.changeDetectionRef.detectChanges();
        });
  }

  refreshGroup(group: Group): void {
    this.group = group;

    // set readonly status depending of current user (but only if a group is really selected)
    if (!this.group) {
      return;
    }

    this.readonly = this.launchInfo.user_id !== this.group.group_leader_id;
  }

  onTabChange(event: NgbTabChangeEvent): void {

    // if we are already on requested tab, do nothing (event fires when tab is selected programmatically)
    if (this.tabRequest && this.tabRequest === event.nextId) {
      return;
    }

    // require confirmation before proceeding if components on active tab are dirty
        if ((this.contract && this.contract.isDirty())
        || (this.roleDefinitions && this.roleDefinitions.isDirty())) {
          this.confirmTabRequest(event);
    }
  }

  confirmTabRequest(event: NgbTabChangeEvent): void {
    event.preventDefault();
    this.tabRequest = event.nextId;
    this.confirmDeleteModal.open(
      'Continue',
      'You have unsaved changes that will disappear if you proceed. Do you want to continue?',
      'Yes',
      'Cancel'
    );
  }

  proceedTabRequest(): void {
    this.tabs.select(this.tabRequest);
    this.clearTabRequest();
  }

  clearTabRequest(): void {
    this.tabRequest = null;
  }
}

