import { Component, Input, Output, OnInit, SimpleChange, SimpleChanges, EventEmitter, ViewChild  } from '@angular/core';
import { Observable, forkJoin } from 'rxjs';
import { map } from 'rxjs/operators';

import { Group } from '../shared/models/group';
import { GroupService } from '../shared/services/group.service';
import { Verification } from '../shared/models/verification';

import { UserApproval } from './user-approval';

import { ConfirmDeleteModalComponent } from '../confirm-delete-modal/confirm-delete-modal.component';

@Component({
  selector: 'app-approval',
  templateUrl: './approval.component.html',
  styleUrls: ['./approval.component.css']
})
export class ApprovalComponent implements OnInit {

  @Input() readonly: boolean;
  @Input() disabled: boolean;
  @Input() group: Group;
  @Input() userId: number;
  @Input() userIsInstructor: boolean;

  @Output() groupStatusChanged: EventEmitter<Group> = new EventEmitter<Group>();

  @ViewChild('confirmUnlockModal') confirmUnlockModal: ConfirmDeleteModalComponent;

  approvals: UserApproval [];
  approvalCurrentUser: UserApproval;
  verifications: Verification [];

  approvalChecked: boolean;
  verifiedChecked: boolean;

  // used for hiding zero-dates (approvals and verifications created before dates were added to database)
  minDate = new Date(2000, 0, 1);

  constructor(private groupService: GroupService) { }

  ngOnInit() {
  }

  ngOnChanges(changes: SimpleChanges) {
    if (changes.group) {
      const groupChange: SimpleChange = changes.group;
      this.setGroup(groupChange.currentValue);
    }
  }

  setGroup(group: Group) {
    this.group = group;

    this.approvals = null;
    this.approvalCurrentUser = null;

    this.getApprovals()
    .subscribe(result => {
      this.approvals = result;

      const approvalCurrentUser = this.approvals.filter(x => x.user.id === this.userId);
      this.approvalCurrentUser = (approvalCurrentUser.length > 0) ? approvalCurrentUser[0] : null;

      this.approvalChecked = false;
    });

    this.groupService.getVerifications(this.group.id)
      .subscribe(result => {
          this.verifications = result;
          this.verifiedChecked = false;
      });
  }

  getApprovals(): Observable<UserApproval[]>  {
    if (!this.group) {
      return;
    }

    return forkJoin(this.groupService.getGroupUsers(this.group.id), this.groupService.getApprovals(this.group.id))
    .pipe(
      map(result => {

        const users = result[0];
        const approvals = [];

          // check if user has approved and add to list
          users.forEach(user => {
            const approval = new UserApproval();
            approval.user = user;
            approval.hasApproved = result[1].filter(x => x.user_id === user.id).length > 0;

            if (approval.hasApproved) {
              approval.approveDate = result[1].filter(x => x.user_id === user.id)[0].approve_date;
            }
            approvals.push(approval);
          });

          return approvals;
    }));
  }


  approve(): void {
    this.groupService.addApproval(this.group.id, this.userId)
    .subscribe((result) => {
      this.approvalCurrentUser.hasApproved = true;
      this.approvalCurrentUser.approveDate = result.approve_date;
      this.approvalChecked = false;
    });
  }

  verify(): void {
    const verification = new Verification();
    verification.user_id = this.userId;

    this.groupService.addVerification(this.group.id, verification)
      .subscribe(result => {
        this.verifications[0] = result;
        this.verifiedChecked = false;
      });
  }

  allMembersHaveApproved(): boolean {
    return this.approvals.filter(x => !x.hasApproved).length === 0;
  }

  someMembersHaveApproved(): boolean {
    return this.approvals.filter(x => x.hasApproved).length > 0;
  }

  unlockClicked(): void {
    if (this.someMembersHaveApproved() || this.verifications.length > 0) {
      this.confirmUnlockModal.open(
        'Unlock',
        'By unlocking you will reset all approvals and the teachers verification. Do you want to continue?',
        'Yes',
        'Cancel'
      );
    } else {
      this.unlock();
    }
  }

  lock(): void {
    this.groupService.saveLockedStatus(this.group.id, true)
      .subscribe(() => {
        this.group.locked = true;
        this.groupStatusChanged.emit(this.group);
      });
  }

  unlock(): void {
    this.groupService.saveLockedStatus(this.group.id, false)
      .subscribe(() => {
        this.approvals.forEach(approval => {
          approval.hasApproved = false;
          approval.approveDate = null;
        });

        this.verifications = [];
        this.group.locked = false;
        this.groupStatusChanged.emit(this.group);
      });
  }
}
