import { Component, Input, OnInit, SimpleChanges, SimpleChange, ViewChild } from '@angular/core';
import { NgForm } from '@angular/forms';

import { Group } from '../shared/models/group';
import { RoleDefinition } from '../shared/models/role-definition';
import { GroupService } from '../shared/services/group.service';

@Component({
  selector: 'app-role-definitions',
  templateUrl: './role-definitions.component.html',
  styleUrls: ['./role-definitions.component.css']
})
export class RoleDefinitionsComponent implements OnInit {

  @Input() readonly: boolean;
  @Input() disabled: boolean;
  @Input() group: Group;

  @ViewChild('editForm') editForm: NgForm;
  @ViewChild('confirmDeleteModal') confirmDeleteModal: any;
  @ViewChild('unsavedChangesModal') unsavedChangesModal: any;

  roles: RoleDefinition[];
  savedRoles: RoleDefinition[];
  selectedRole: RoleDefinition;
  roleRequest: RoleDefinition;

  // variables to handle handle state and validation of ngb-rating
  errorMessageWeight: string;
  weightTouched: boolean;
  weightDirty: boolean;

  constructor(private groupService: GroupService) { }

  ngOnInit() {}

  ngOnChanges(changes: SimpleChanges) {
    if (changes.group) {
      const groupChange: SimpleChange = changes.group;
      this.setGroup(groupChange.currentValue);
    }
  }

  setGroup(group: Group) {
    this.group = group;

    this.roles = null;
    this.getRoles();
  }

  getRoles(): void {

    if (this.group) {
      this.groupService.getRoleDefinitions(this.group.id)
        .subscribe(result => {
          this.roles = result;
          this.setSavedRoles();
        });
    }
  }

  roleClicked(clickedRole: RoleDefinition): void {

    // if role is already selected do nothing
    if (this.selectedRole && this.selectedRole.id === clickedRole.id) {
      return;
    }

    // find requested role in saved list
    let requestedRole: RoleDefinition;
    if (this.roles.filter(x => x.id === clickedRole.id).length > 0) {
      requestedRole = this.roles.filter(x => x.id === clickedRole.id)[0];
    }

    // require confirmation before proceeding if there are unsaved changes
    if (this.isDirty()) {
        this.openUnsavedChangesModal(requestedRole);
    } else {
      this.setSelectedRole(requestedRole);
    }
  }

  openUnsavedChangesModal(role: RoleDefinition): void {
    this.roleRequest = role;
    this.unsavedChangesModal.open(
      "Unsaved changes",
      "You have unsaved changes that will disappear if you proceed. Do you want to continue?",
      "Yes",
      "Cancel"
    );
  }

  proceedRoleRequest(): void {
    this.resetRolesToSavedState();
    this.setSelectedRole(this.roleRequest);
    this.roleRequest = null;
  }

  cancelRoleRequest(): void {
    this.roleRequest = null;
  }


  cleanForm(): void {
    if (this.editForm) {
      this.editForm.form.markAsPristine();
      this.editForm.form.markAsUntouched();
    }
    this.weightTouched = false;
    this.weightDirty = false;
  }

  setSelectedRole(role: RoleDefinition): void {
    this.cleanForm();
    this.selectedRole = role;
  }

  resetRolesToSavedState(): void {
    this.cleanForm();
    this.roles = JSON.parse(JSON.stringify(this.savedRoles));
  }

  setSavedRoles(): void {
    this.savedRoles = JSON.parse(JSON.stringify(this.roles));
    this.selectedRole = null;
  }

  createNewRole(): void {
    const newRole = new RoleDefinition();
    newRole.weight = 0;

    // check if there are unsaved changes before selecting new role
    if (this.isDirty()) {
      this.openUnsavedChangesModal(newRole);
    } else {
      this.setSelectedRole(newRole);
    }

    this.validateWeight();
  }

  saveSelectedRole(): void {

    // check saved list to determine if we're updating existing role or adding a new
    if (this.savedRoles.filter(x => x.id === this.selectedRole.id).length === 0) {
      this.groupService.addRoleDefinition(this.group.id, this.selectedRole)
      .subscribe(result => {

        // add new role to list and update saved list
        this.roles.push(result);
        this.setSavedRoles();
      });
    } else {
      this.groupService.updateRoleDefinition(this.group.id, this.selectedRole)
      .subscribe(() => {
        // update saved list
        this.setSavedRoles();
      });
    }
  }

  deleteSelectedRole(confirm: boolean = true) {

    if (confirm) {
      this.confirmDeleteModal.open("Remove role", `Are the sure you want to remove the role ${this.selectedRole.name}?`, "Yes", "Cancel");
    } else {
      this.groupService.deleteRoleDefinition(this.group.id, this.selectedRole)
      .subscribe(() => {
          // remove deleted role from list and update saved list
          this.roles = this.roles.filter(x => x.id !== this.selectedRole.id);
          this.setSavedRoles();
      });
    }
  }

  isDirty(): boolean {
    return this.editForm && this.editForm.form.dirty;
  }

  // handle ngb-rating not setting ng-touched and ng-dirty of form and validation of weight
  weightLeaved(): void {

    this.weightTouched = true;

    if (this.editForm) {
      this.editForm.form.markAsTouched();
    }

    this.validateWeight();
  }

  weightChanged(): void {

    this.weightDirty = true;

    if (this.editForm) {
      this.editForm.form.markAsDirty();
    }

    this.validateWeight();
  }

  validateWeight(): void {

    if (this.selectedRole.weight < 1) {
      this.errorMessageWeight = "Expected workload must be stated.";
    } else {
      this.errorMessageWeight = null;
    }
  }
}
