import { Component, Input, OnInit, OnChanges, SimpleChange, SimpleChanges  } from '@angular/core';
import { forkJoin } from 'rxjs';

import { Group } from '../shared/models/group';
import { Role } from '../shared/models/role';
import { User } from '../shared/models/user';
import { RoleDefinition } from '../shared/models/role-definition';

import { GroupService } from '../shared/services/group.service';
import { MessageService } from '../shared/services/message.service';

@Component({
  selector: 'app-roles',
  templateUrl: './roles.component.html',
  styleUrls: ['./roles.component.css']
})

export class RolesComponent implements OnInit, OnChanges {

  @Input() readonly: boolean;
  @Input() disabled: boolean;
  @Input() group: Group;

  users: User[];
  roles: Role[];
  roledefinitions: RoleDefinition[];
  userWeights: Map<number, number> = new Map<number, number>();
  selectedRoleDefinition: RoleDefinition;
  selectedUser: User;

  constructor(private groupService: GroupService, private messageService: MessageService) { }

  ngOnInit() {}

  ngOnChanges(changes: SimpleChanges) {
    if (changes.group) {
      const groupChange: SimpleChange = changes.group;
      this.setGroup(groupChange.currentValue);
    }
  }

  setGroup(group: Group) {
    this.group = group;

    this.users = null;
    this.roles = null;
    this.roledefinitions = null;
    this.userWeights.clear();

    this.getUsersRolesAndRoledefinitions();
  }

  getUsersRolesAndRoledefinitions(): void {
    if (!this.group) {
      return;
    }

    forkJoin([
      this.groupService.getGroupUsers(this.group.id),
      this.groupService.getGroupRoles(this.group.id),
      this.groupService.getRoleDefinitions(this.group.id)
    ]).subscribe(([resp1, resp2, resp3]) => {
      this.users = resp1;
      this.roles = resp2;
      this.roledefinitions = resp3;
      this.users.forEach(user => this.userWeights.set(user.id, this.getWeightForUser(user.id)));
    });
  }

  getRolesForUser(userId: number): RoleDefinition[] {
    if (!this.roles) {
      return [];
    }

    const roledefs = [];
    this.roles.filter(x => x.user_id === userId).forEach(r => {
      this.roledefinitions.filter(y => y.id === r.role_definition_id).forEach(rd => {
          roledefs.push(rd);
      });
    });

    return roledefs;
  }

  getWeightForUser(userId: number): number {
    if (!this.roles) {
      return 0;
    }

    let weight = 0;
    this.roles.filter(x => x.user_id === userId).forEach(r => {
      this.roledefinitions.filter(y => y.id === r.role_definition_id).forEach(rd => {
        weight += rd.weight;
      });
    });

    return weight;
  }

  isRoleDefinitionAssigned(roleDefId: number): boolean {
    return (this.roles.findIndex(x => x.role_definition_id === roleDefId) !== -1);
  }

  userHasRole(userId: number, roledefId: number): boolean {
    return this.roles.findIndex(x => x.user_id === userId && x.role_definition_id === roledefId) !== -1;
  }

  deleteRoleFromUser(userId: number, roledefId: number): void {

    if (this.disabled) {
      return;
    }

    if (!this.userHasRole(userId, roledefId )) {
      return;
    }

    const idx = this.roles.findIndex(x => x.user_id === userId && x.role_definition_id === roledefId);

    this.groupService.deleteGroupRole(this.group.id, this.roles[idx].id)
      .subscribe({
        next: () => {
          this.roles.splice(idx, 1);
          this.userWeights.set(userId, this.getWeightForUser(userId));
         }
      });
  }

  addRoleToUser(user: User, roledef: RoleDefinition): void {

    if (this.userHasRole(user.id, roledef.id )) {
      return;
    }

    this.groupService.addGroupRole(this.group.id, user, roledef)
      .subscribe({
        next: resp => {
          this.roles.push(resp);
          this.userWeights.set(user.id, this.getWeightForUser(user.id));

          this.selectedRoleDefinition = null;
          this.selectedUser = null;
         }
      });
  }

  assignRoleAutomatically(roledef: RoleDefinition): void {
    const users_by_weight = this.users.filter(u => !this.userHasRole(u.id, roledef.id)).map(u => {
      return {
        user: u,
        weight: this.getWeightForUser(u.id)
      };
    }).sort((a, b) => a.weight < b.weight ? -1 : a.weight === b.weight ? 0 : 1);

    const candidates = users_by_weight.filter(u => u.weight === users_by_weight[0].weight);
    if (candidates.length === 0) {
      this.messageService.addWarning('No group member can be assigned the role.');
      return;
    }
    const random_idx = Math.floor(Math.random() * candidates.length);

    this.addRoleToUser(candidates[random_idx].user, roledef);
  }

  selectUser(user: User): void {
    if (this.selectedUser === user) {
      this.selectedUser = null;
    } else {
      this.selectedUser = user;
    }
  }

  selectRoleDefinition(roledef: RoleDefinition): void {
    if (this.selectedRoleDefinition === roledef) {
      this.selectedRoleDefinition = null;
    } else {
      this.selectedRoleDefinition = roledef;
    }
  }
}
