import { Injectable, Inject } from '@angular/core';
import { HttpClient, HttpHeaders } from '@angular/common/http';

import { Observable, BehaviorSubject, of, throwError } from 'rxjs';
import { catchError, tap, map } from 'rxjs/operators';

import { environment } from '../../../environments/environment';

import { Group } from '../models/group';
import { Contract } from '../models/contract';
import { Role } from '../models/role';
import { RoleDefinition } from '../models/role-definition';
import { User } from '../models/user';
import { LaunchInfo } from '../models/launch-info';
import { Approval } from '../models/approval';
import { EducationalResource } from '../models/educational-resource';
import { Verification } from '../models/verification';

import { MessageService } from './message.service';

const httpOptions = {
  headers: new HttpHeaders({ 'Content-Type': 'application/json' })
};

@Injectable({
  providedIn: 'root'
})
export class GroupService {

  private groupUrl = `${environment.apiUrl}/groups`;

  constructor(
    private http: HttpClient,
    private messageService: MessageService,
    @Inject('CGLSessionToken') sessionToken: BehaviorSubject<string>
  ) {
    sessionToken.subscribe((token) => httpOptions.headers = httpOptions.headers.set('Authorization', 'CGL ' + token));
  }

  /** GET info about current launch (current course, user, users role in course) */
  getLaunchInfo(): Observable<LaunchInfo> {
    return this.http.get<LaunchInfo>(`${environment.apiUrl}/ltiinfo`, httpOptions);
  }

  /** GET all groups (for current user on current course) */
  getGroups(): Observable<Group []> {
    return this.http.get<Group[]>(this.groupUrl, httpOptions).pipe(
      catchError(this.handleBreakingError<Group[]>('Could not fetch the groups.'))
    )
  }

  /** GET educational resources */
  getEducationalResources(): Observable<EducationalResource []> {
    const url = `${environment.apiUrl}/educationalResources`;

    return this.http.get<EducationalResource[]>(url, httpOptions)
  }

  /** GET roles of members in specified group */
  getGroupRoles(groupId: number): Observable<Role []> {
    const url = `${this.groupUrl}/${groupId}/roles`;

    return this.http.get<Role[]>(url, httpOptions);
  }

  /** POST new role for member in specified group */
  addGroupRole(groupId: number, user: User, roledef: RoleDefinition): Observable<Role> {
    const url = `${this.groupUrl}/${groupId}/roles`;

    let role: Role = {
      id: 0,
      role_definition_id: roledef.id,
      user_id: user.id
    };

    return this.http.post<Role>(url, role, httpOptions).pipe(
      tap(_ => this.messageService.addInfo(`The role ${roledef.name} was assigned to ${user.name}.`)),
      catchError(this.handleBreakingError<Role>('The assigned role could not be saved.'))
    );
  }

  /** DELETE role for member in specified group */
  deleteGroupRole(groupId: number, roleId: number): Observable<any> {
    const url = `${this.groupUrl}/${groupId}/roles/${roleId}`;

    return this.http.delete(url, httpOptions).pipe(
      tap(_ => this.messageService.addInfo('The role assignment is removed.')),
      catchError(this.handleBreakingError<Role>('The assigned role could not be removed.'))
    );
  }

  /** GET members of specified group */
  getGroupUsers(groupId: number): Observable<User []> {
    const url = `${this.groupUrl}/${groupId}/users`;

    return this.http.get<User[]>(url, httpOptions);
  }

  /** GET contract for specified group */
  getContract (groupId: number): Observable<Contract> {

    const url = `${this.groupUrl}/${groupId}/contract`;

    return this.http.get<Contract>(url, httpOptions);
  }

  /** PUT contract for specified group */
  updateContract(groupId: number, contract: Contract): Observable<any> {
    const url = `${this.groupUrl}/${groupId}/contract`;

    return this.http.put<Contract>(url, contract, httpOptions).pipe(
      tap(_ => this.messageService.addInfo('Changes have been saved.')),
      catchError(this.handleBreakingError<Contract>('Could not save changes.'))
    );
  }

  /** GET standard contract */
  getStandardContract(): Observable<Contract> {
    const url = `${this.groupUrl}/standardContract`;

    return this.http.get<Contract>(url, httpOptions).pipe(
      catchError(this.handleBreakingError<Contract>('Could not fetch contract template.'))
    );
  }

  /** GET role definitions for specified group */
  getRoleDefinitions(groupId: number): Observable<RoleDefinition[]>{
    const url = `${this.groupUrl}/${groupId}/roleDefinitions`;

    return this.http.get<RoleDefinition[]>(url, httpOptions);
  }

  /** POST new role definition for specified group */
  addRoleDefinition(groupId: number, roledefinition: RoleDefinition): Observable<RoleDefinition> {
    const url = `${this.groupUrl}/${groupId}/roleDefinitions`;

    return this.http.post<RoleDefinition>(url, roledefinition, httpOptions).pipe(
      tap(_ => this.messageService.addInfo(`The role ${roledefinition.name} has been added.`)),
      catchError(this.handleBreakingError<RoleDefinition>('The role could not be added.'))
    )
  }

  /** PUT updated role definition for specified group */
  updateRoleDefinition(groupId: number, roledefinition: RoleDefinition): Observable<any> {
    const url = `${this.groupUrl}/${groupId}/roleDefinitions/${roledefinition.id}`;

    return this.http.put<RoleDefinition>(url, roledefinition, httpOptions).pipe(
      tap(_ => this.messageService.addInfo('Changes have been saved.')),
      catchError(this.handleBreakingError<RoleDefinition>('Could not save changes.'))
    );
  }

  /** DELETE role definition for specified group */
  deleteRoleDefinition(groupId: number, roledefinition: RoleDefinition): Observable<any> {
    const url = `${this.groupUrl}/${groupId}/roleDefinitions/${roledefinition.id}`;

    return this.http.delete<RoleDefinition>(url, httpOptions).pipe(
      tap(_ => this.messageService.addInfo(`The role ${roledefinition.name} has been removed.`)),
      catchError(this.handleBreakingError<RoleDefinition>('The role could not be removed.'))
    );
  }


  /** GET members who have approved contract and roles of specified group */
  getApprovals(groupId: number): Observable<Approval[]> {
    const url = `${this.groupUrl}/${groupId}/approvals`;

    return this.http.get<Approval[]>(url, httpOptions).pipe(
      map(approvals => {
        approvals.forEach(approval => {
          approval.approve_date = new Date(approval.approve_date);
        });

        return approvals;
      })
    );
  }

  /** POST a new approval for a specified group */
  addApproval(groupId: number, userId: number): Observable<Approval> {

    const url = `${this.groupUrl}/${groupId}/approvals`;

    let approval = new Approval();
    approval.user_id = userId;

    return this.http.post<Approval>(url, approval, httpOptions).pipe(
      tap(_ => this.messageService.addInfo('Your approval is saved.')),
      map(approval => {
        approval.approve_date = new Date(approval.approve_date);
        return approval;
      }),
      catchError(this.handleBreakingError<Approval>('Your approval could not be saved.'))
    );
  }

  /** GET teachers who have verified contract and roles of specified group */
  getVerifications(groupId: number): Observable<Verification[]> {
    const url = `${this.groupUrl}/${groupId}/verifications`;

    return this.http.get<Verification[]>(url, httpOptions).pipe(
      map(verifications => {

        verifications.forEach(verification => {
          verification.verify_date = new Date(verification.verify_date);
        });

        return verifications;
      })
    );
  }


  /** POST a new verification for specified group */
  addVerification (groupId: number, verification: Verification): Observable<any> {
    const url = `${this.groupUrl}/${groupId}/verifications`;

    return this.http.post<Verification>(url, verification, httpOptions).pipe(
      tap(_ => this.messageService.addInfo('The verification is saved.')),
      map(verification => {
        verification.verify_date = new Date(verification.verify_date);

        return verification;
      }),
      catchError(this.handleBreakingError<Verification>('The verification could not be saved.'))
    );
  }

  /** PUT group locked status */
  saveLockedStatus(groupId: number, locked: boolean): Observable<any> {
    const url = `${this.groupUrl}/${groupId}/locked`;

    return this.http.put<boolean>(url, locked, httpOptions).pipe(
      tap(_ => {
        if (locked) {
          this.messageService.addInfo('Contract and roles are locked for editing.');
        } else {
          this.messageService.addInfo('Contract and roles are unlocked for editing.');
        }
      }),
      catchError(this.handleBreakingError<Approval>('The change could not be saved.'))
    );
  }

  /**
   * Handle Http operation that failed, let the app continue silently.
   * @param result - optional value to return as the observable result
   */
  private handleNonBreakingError<T> (result?: T) {
    return (error: any): Observable<T> => {

      // console.error(error);

      // Let the app keep running by returning an empty result.
      return of(result as T);
    };
  }

  /**
   * Handle Http operation that failed, when app should not continue silently.
   * @param userMessage - a message to present to the user
   */
  private handleBreakingError<T> (userMessage: string) {
    return (error: any): Observable<T> => {

      // console.error(error);

      this.messageService.addError(userMessage);

      return throwError(error);
    };
  }

}
